package com.test.log;

import java.io.Serializable;
import java.util.EmptyStackException;
import java.util.Stack;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class TimeLogger {

	/**
	 * Comment for <code>serialVersionUID</code>
	 */
	private static final long serialVersionUID = -324150805566444828L;
	private static Log log = LogFactory.getLog(TimeLogger.class);
	// 5 * 60 min max age
	private static final long maxAge = 1000 * 60 * 60 * 5;

	private static ThreadLocal<Stack<TimeEntry>> buffer = new ThreadLocal<Stack<TimeEntry>>() {

		@Override
		protected Stack<TimeEntry> initialValue() {
			return new Stack<TimeEntry>();
		}
	};

	public static void beginLog(Class<? extends Object> source,
			String actionDescription) {
		beginLog(source, actionDescription, false, true);
	}

	private static void beginLog(Class<? extends Object> source,
			String actionDescription, boolean useSourceClassLogger,
			boolean withoutLogOutput) {
		if (log.isInfoEnabled()) {
			TimeEntry entry = new TimeEntry(source,
					actionDescription, useSourceClassLogger);
			buffer.get().push(entry);
			
			StringBuilder sb = new StringBuilder();
			if (source != null && !useSourceClassLogger) {
				sb.append(source.getSimpleName());
				sb.append(".");
			}
			
			sb.append(actionDescription);
			sb.append(" started at ").append(entry.startTime);
			if (source != null && useSourceClassLogger) {
				LogFactory.getLog(source).info(sb.toString());
			}
			if (!withoutLogOutput) {
				log.info(sb.toString());
			}
		}
	}

	public static void endLog() {
		endLog(false);
	}

	private static void endLog(boolean useSourceClassLogger) {
		long endTime = System.currentTimeMillis();
		TimeEntry entry = getNextEntry();
		if (entry == null) {
			return;
		}
		printEndStatement(entry, useSourceClassLogger, endTime);
		entry.setEnded(true);
		removeStaleEntries();
	}

	private static void printEndStatement(TimeEntry entry,
			boolean useSourceClassLogger, long endTime) {
		long duration = endTime - entry.getStartTime();
		StringBuilder sb = new StringBuilder();
		if (entry.getSource() != null && !useSourceClassLogger) {
			sb.append(entry.getSource().getSimpleName());
			sb.append(".");
		}
		sb.append(entry.getActionDescription());
		sb.append(" completed at ").append(endTime).append(" after ")
				.append(duration).append(" ms");
		if (entry.getSource() != null
				&& (useSourceClassLogger || entry.isUseSourceClassLogger())) {
			LogFactory.getLog(entry.getSource()).info(sb.toString());
		}
		if (log.isInfoEnabled()) {
			log.info(sb.toString());
		}
	}

	private static TimeEntry getNextEntry() {
		if (buffer == null) {
			log.warn("internal buffer is null");
			return null;
		}
		try {
			if (buffer.get() != null) {
				return (TimeEntry) buffer.get().pop();
			}
		} catch (EmptyStackException e) {
			log.warn("stack is empty. called end without begin or entry was removed after maxAge "
					+ maxAge + " passed.");
		}
		return null;
	}

	private static void removeStaleEntries() {
		TimeEntry entry = null;
		Stack<TimeEntry> stack = buffer.get();
		Stack<TimeEntry> checkedStack = new Stack<TimeEntry>();
		long currentTime = System.currentTimeMillis();
		try {
			while ((entry = stack.pop()) != null) {
				if (!entry.ended && (currentTime - entry.startTime) < maxAge) {
					checkedStack.push(entry);
				} else {
					log.warn("removed stale entry " + entry + " after "
							+ maxAge + " elapsed.");
				}
			}
		} catch (EmptyStackException e) {
		}
		try {
			while ((entry = checkedStack.pop()) != null) {
				stack.push(entry);
			}
		} catch (EmptyStackException e) {
		}
	}

	static class TimeEntry implements Serializable {

		/**
		 * Comment for <code>serialVersionUID</code>
		 */
		private static final long serialVersionUID = -631974753405232096L;
		private long startTime = System.currentTimeMillis();
		private String actionDescription;
		private Class<? extends Object> source;
		private boolean useSourceClassLogger;
		private boolean ended;

		public TimeEntry(Class<? extends Object> source,
				String logmessage, boolean useSourceClassLogger) {
			super();
			this.actionDescription = logmessage;
			this.source = source;
			this.useSourceClassLogger = useSourceClassLogger;
		}

		public long getStartTime() {
			return startTime;
		}

		public void setStartTime(long startTime) {
			this.startTime = startTime;
		}

		public String getActionDescription() {
			return actionDescription;
		}

		public void setActionDescription(String logmessage) {
			this.actionDescription = logmessage;
		}

		public Class<? extends Object> getSource() {
			return source;
		}

		public void setSource(Class<? extends Object> source) {
			this.source = source;
		}

		public boolean isUseSourceClassLogger() {
			return useSourceClassLogger;
		}

		public void setUseSourceClassLogger(boolean useSourceClassLogger) {
			this.useSourceClassLogger = useSourceClassLogger;
		}

		@Override
		protected void finalize() throws Throwable {
			if (!ended) {
				StringBuilder sb = new StringBuilder();
				sb.append("PerformanceLogger was not ended for action ")
						.append(getActionDescription()).append(" in Class ")
						.append(getSource());
				log.warn(sb.toString());
			}
		}

		public boolean isEnded() {
			return ended;
		}

		public void setEnded(boolean ended) {
			this.ended = ended;
		}

		@Override
		public String toString() {
			StringBuilder sb = new StringBuilder();
			sb.append("PerformanceEntry[");
			sb.append("class=");
			if (getSource() != null) {
				sb.append(getSource().getName());
			}
			sb.append(", start=").append(getStartTime());
			sb.append(", message=").append(getActionDescription());
			sb.append("]");
			return sb.toString();
		}

	}

}
